#include <ncbi.h>
#include <readdb.h>
#include "psublast/libc.h"
#include "psublast/psublast.h"
#include "db.h"
#include "mb.h"

#ifndef __lint
static const char rcsid[] =
"$Id: mbmain.c,v 1.36 1999/09/24 19:04:35 schwartz Exp $";
#endif

#define NUMARG (sizeof(myargs)/sizeof(myargs[0]))
static Args myargs[] =
{
    {"Database", /*0*/
     NULL, NULL, NULL, FALSE, 'd', ARG_STRING, 0.0, 0, NULL},
    {"Query File", /*1*/
     "stdin", NULL, NULL, FALSE, 'i', ARG_FILE_IN, 0.0, 0, NULL},
    {"Dropoff (X) for extensions", /*2*/
     "180", NULL, NULL, FALSE, 'X', ARG_INT, 0.0, 0, NULL},
    {"Length of perfect match (in bytes) before extension", /*3*/
     "7", NULL, NULL, FALSE, 'w', ARG_INT, 0.0, 0, NULL},
    {"Maximal number of entries allowed in a 12mer", /*4*/
     "100", NULL, NULL, FALSE, 'p', ARG_INT, 0.0, 0, NULL},
    {"Number of charaters output for sequence id, maximum 50", /*5*/
     "40", NULL, NULL, FALSE, 'n', ARG_INT, 0.0, 0, NULL},
    {"Perform est search instead of blast search (T/F)", /*6*/
     "F", NULL, NULL, FALSE, 'e', ARG_BOOLEAN, 0.0, 0, NULL},
    {"Usage", /*7*/
     "F", NULL, NULL, FALSE, '?', ARG_BOOLEAN, 0.0, 0, NULL},
    {"Mismatch penalty (greedy extension) (match is +10)", /*8*/
     "-20", NULL, NULL, FALSE, 'r', ARG_INT, 0.0, 0, NULL},
    {"Extension option 0:Greedy 1:blastz 2:Greedy affine", /*9*/
     "0", NULL, NULL, FALSE, 'P', ARG_INT, 0.0, 0, NULL},
    {"Match score (0 for defaults)",  /*10*/
     "10", NULL, NULL, FALSE, 'M', ARG_INT, 0.0, 0, NULL},
    {"Mismatch penalty (blastz) (0 for defaults)", /*11*/
     "0", NULL, NULL, FALSE, 'R', ARG_INT, 0.0, 0, NULL},
    {"GapOpen Penalty (0 for defaults)",  /*12*/
     "0", NULL, NULL, FALSE, 'O', ARG_INT, 0.0, 0, NULL},
    {"Gap Extension Penalty (0 for defaults)", /*13*/
     "0", NULL, NULL, FALSE, 'E', ARG_INT, 0.0, 0, NULL},
    {"Threshold to output blastz alignment", /*14*/
     "240", NULL, NULL, FALSE, 'T', ARG_INT, 0.0, 0, NULL},
    {"Cutoff for gap-alignment",  /*15*/
     "200", NULL, NULL, FALSE, 't', ARG_INT, 0.0, 0, NULL},
    {"X dropoff for gapfree extension",  /*16*/
     "50", NULL, NULL, FALSE, 'x', ARG_INT, 0.0, 0, NULL},
    {"Number of words output for sequence id",  /*17*/
     "0", NULL, NULL, FALSE, 'N', ARG_INT, 0.0, 0, NULL},
    {"Normalize the coordinates of reverse compliment output" ,  /*18*/
     "F", NULL, NULL, FALSE, 'Z', ARG_BOOLEAN, 0.0, 0, NULL},
    {"Output alignment mode: (0) no alignment (1) summary (2) alignment",  /*19*/
     "0", NULL, NULL, FALSE, 'A', ARG_INT, 0.0, 0, NULL},
    {"Query reverse compliment or not",  /*20*/
     "T", NULL, NULL, FALSE, 'c', ARG_BOOLEAN, 0.0, 0, NULL},
};

static void usage()
{
    fprintf(stderr, "usage:\n");
    fprintf(stderr, " -i query file\n");
    fprintf(stderr, " -d database file\n");
    fprintf(stderr, " -x dropoff for extensions\n");
    fprintf(stderr, " -w length of perfect match before extension\n");
    fprintf(stderr, " -p maximal number of entries allowed in a 12mer\n");
    fprintf(stderr, " -n width of printed sequence id\n");
    fprintf(stderr, " -e est flag\n");
    exit(1);
}

static void readdb_get_stuff(ReadDBFILEPtr rdpt, int dbidx,
		uchar **s, int *slen, char **defline)
{
    SeqIdPtr sip = 0;
    *slen = readdb_get_sequence(rdpt, dbidx, s);
    readdb_get_descriptor(rdpt, dbidx, &sip, defline);
    sip = SeqIdSetFree(sip);
}

Int2 Main(void)
{
    char *blast_database;
    char *blast_inputfile;
    FILE *infp;
    ReadDBFILEPtr rdpt;
    mbt_t *mbt;
    qdb_t *qdb;
    int i;
    align_params_t P;
    int revcomp_flag;

    if (!GetArgs("mb", NUMARG, myargs))
	return (1);
    
    blast_database = myargs[0].strvalue;
    blast_inputfile = myargs[1].strvalue;
    P.xdrop_ext = myargs[2].intvalue;
    P.len_perf_match = myargs[3].intvalue-1;
    P.max_12mer_ents = myargs[4].intvalue;

    P.idlen = myargs[5].intvalue;
    i = myargs[17].intvalue;
    if (i > 0)
	P.idlen = -i;

    P.mismatch_penalty_mb = myargs[8].intvalue;
    if (P.mismatch_penalty_mb >= 0) fatal("mismatch must be negative");
    P.match_score = myargs[10].intvalue;
    P.mismatch_penalty_bz = myargs[11].intvalue;
    P.gap_open_penalty = myargs[12].intvalue;
    P.gap_extension_penalty = myargs[13].intvalue;
    P.bzthresh = myargs[14].intvalue;
    P.gap_align_cutoff = myargs[15].intvalue;
    P.xdrop_gapfree = myargs[16].intvalue;
    P.normalize_revcomp_flag = myargs[18].intvalue;

    /* prog_flag = 0: est_search 1: mb with greedy 2: mb with blastz*/
    P.prog_flag = myargs[6].intvalue ? MB_DO_EST :
		  ((myargs[9].intvalue==1) ? MB_DO_BZ : MB_DO_GREEDY);
    P.output_al =  myargs[19].intvalue;
    revcomp_flag = myargs[20].intvalue;

    if (myargs[9].intvalue == 0) { /* greedy case */
        P.match_score = MATCH_SCORE_SCALE;
        P.gap_open_penalty = 0;
        P.gap_extension_penalty = -P.mismatch_penalty_mb + MATCH_SCORE_SCALE/2;
    } else if (myargs[9].intvalue == 1) { /* blastz case */
	/* nothing to do */;
    } else if (myargs[9].intvalue == 2) { /* affine case */
        if (P.gap_extension_penalty <= 0) 
          fatal("In the affine case, the gap extension penalty must be positive.");
    }

    if (P.idlen > HDRSAVE) P.idlen = HDRSAVE;

    if ((infp = FileOpen(blast_inputfile, "r")) == NULL) {
	ErrPostEx(SEV_FATAL, 0, 0,
            "blast: Unable to open input file %s\n", blast_inputfile);
	return (1);
    }

    mbt = mb_init();
    if (P.prog_flag == MB_DO_BZ) {
	fatal("blastz no longer supported!");
    }

    if ((rdpt = readdb_new(blast_database, 0)) == 0)
	fatalf("cannot open database '%s'", blast_database);

    qdb = (qdb_t*) ckalloc(sizeof(*qdb));

    do {
        int nqseq = qdb_fill_block(infp, qdb, revcomp_flag);
        int len = qdb_total_length(qdb);
        int dbidx;

        mb_start(mbt, qdb->sequence, len, P.len_perf_match, P.max_12mer_ents); /* XXX */
	for (dbidx = 0; dbidx < rdpt->num_seqs; ++dbidx) {
            uchar *s;
            int slen;
	    char *d;

	    readdb_get_stuff(rdpt, dbidx, &s, &slen, &d);
	    if (P.prog_flag != MB_DO_EST)
                mb_search(mbt, s,slen, nqseq,qdb, d, &P);
	    else
	        if (est_search(mbt,s,slen,&P))
	            mb_print_sequence(s,slen,d);

            d = MemFree(d);
	}
    } while (!feof(infp) && (P.prog_flag != MB_DO_EST));

    mb_end(mbt);
    FileClose(infp);
    MemFree(qdb);
    readdb_destruct(rdpt);
    return 0;
}
