#ifndef __lint
static const char rcsid[] =
"$Id: blast_filter.c,v 1.13 1999/03/30 21:57:01 schwartz Exp schwartz $";
#endif

#include "psublast/libc.h"
#include "psublast/psublast.h"

/* -----------------------------  filter the MSPs  ---------------------------*/

msp_table_t *msp_compress(msp_table_t *mt)
{
	msp_t *p, *q;
	int i = 0;

	if (mt == 0)
		return 0;
	if (mt->msp == 0)
		return 0;

	for (q=p=MSP_TAB_FIRST(mt); MSP_TAB_MORE(mt,p); p=MSP_TAB_NEXT(p)) {
		if (p->filter == 0) {
			if (p != q)
				*q = *p;
			q = MSP_TAB_NEXT(q);		
			++i;
		} 
	}
	mt->num = i;
	return mt;
}

#define DEFINE_MSP_CMP(f) \
int msp_cmp_##f(const void *a, const void *b) { \
	return ((const msp_t*)a)->f - ((const msp_t*)b)->f; \
}
DEFINE_MSP_CMP(pos1)
DEFINE_MSP_CMP(pos2)

msp_table_t *msp_sort_by(msp_table_t *mt, msp_cmp_t cmp)
{
	msp_t *msp; 
	unsigned int num;

	if ((mt == NULL) || ((num = mt->num) == 0))
		return 0;

        msp = mt->msp;
	qsort((char *) msp, num, sizeof(msp_t), cmp);
	return mt;
}

#define DEFINE_MSP_SORT(f) \
msp_table_t *msp_sort_##f(msp_table_t *mt) { \
      return msp_sort_by(mt, msp_cmp_##f); \
}
DEFINE_MSP_SORT(pos1)
DEFINE_MSP_SORT(pos2)

msp_table_t *msp_filter(msp_table_t *mt)
{
	msp_t *mp, *q;
	int right_most, i;

	if ((mt == NULL) || (mt->msp == 0))
		return 0;

	msp_sort_pos1(mt);
	right_most = mt->msp->pos1 + mt->msp->len; 
	q = MSP_TAB_FIRST(mt);
	for (mp=MSP_TAB_NEXT(q); MSP_TAB_MORE(mt,mp); mp=MSP_TAB_NEXT(mp)) {
		if ((i=mp->pos1 + mp->len) > right_most) {
			if (q->pos1 >= mp->pos1 -10) q->filter = 1;
			if (i <= right_most + 10) mp->filter = 1;
			q = mp; 
			right_most = i;
		} else {
			mp->filter = 1;
			if (i >= right_most -10 && mp->pos1 <= q->pos1+10)
				q->filter = 1;
		}
	}
	msp_sort_pos2(mt);
	right_most = mt->msp->pos2+mt->msp->len; 
	q = MSP_TAB_FIRST(mt);
	for (mp=MSP_TAB_NEXT(q); MSP_TAB_MORE(mt,mp); mp=MSP_TAB_NEXT(mp)) {
		if ((i=mp->pos2 + mp->len) > right_most) {
			if (q->pos2 >= mp->pos2 -10) q->filter = 1;
			if (i <= right_most + 10) mp->filter = 1;
			q = mp; 
			right_most = i;
		} else {
			mp->filter = 1;
			if (i >= right_most -10 && mp->pos2 <= q->pos2+10)
				q->filter = 1;
		}
	}
	return mt;
}

msp_table_t *msp_filter_old(msp_table_t *mt)
{
	int leni, lenj, pos1i, pos2i, pos1j, pos2j;
	msp_t *mp, *q;

	for (mp=MSP_TAB_FIRST(mt); MSP_TAB_MORE(mt,mp); mp=MSP_TAB_NEXT(mp)) {
		/* see which MSPs are knocked out by the i-th MSP */
		pos1i = mp->pos1;
		pos2i = mp->pos2;
		leni = mp->len;
		for (q=MSP_TAB_FIRST(mt); MSP_TAB_MORE(mt,q); q=MSP_TAB_NEXT(q))
			if (q != mp) {
				pos1j = q->pos1;
				pos2j = q->pos2;
				lenj = q->len;
				/* is hit j approx. inside hit i ? */
				/* WRT sequence #1 ..  */
				if (pos1i - 10 <= pos1j &&
				    pos1i + leni + 10 >= pos1j + lenj)
					q->filter = 1;
				/* or WRT sequence #2 */
				if (pos2i - 10 <= pos2j &&
				    pos2i + leni + 10 >= pos2j + lenj)
					q->filter = 1;
			}
	}
	return mt;
}

/* --------------------------  sort the MSPs  ------------------------------- */

/* sort MSPs by score */
int msp_cmp_score(const void *e, const void *f)
{
	int i;
	const msp_t *ee = (const msp_t *)e;
	const msp_t *ff = (const msp_t *)f;
	

	if ((i = (ee->score - ff->score)) < 0)
                return 1;
        else if (i > 0)
                return -1;
        else if ((i = (ee->pos1 - ff->pos1)) > 0)
                return 1;
	else if (i < 0)
		return -1;
        else if ((i = (ee->pos2 - ff->pos2)) > 0)
                return 1;
	else if (i < 0)
		return -1;
	else
                return 0;
} 

msp_table_t *msp_sort(msp_table_t *mt)
{
	return msp_sort_by(mt, msp_cmp_score);
}


/* -----------------------------  print the MSPs  ----------------------------*/

static int percent_match(SEQ *sf1, SEQ *sf2, int len, int pos1, int pos2)
{
	uchar *p = SEQ_CHARS(sf1)+pos1; /* XXX - +1? */
	uchar *q = SEQ_CHARS(sf2)+pos2; /* XXX - +1? */
	int m = 0;
	int i = len;
	while (i-- > 0)
		m += (*p++ == *q++);
	return align_match_percent(len, m);
}

void msp_print(SEQ *sf1, SEQ *sf2, struct msp_table *mt)
{
	int len, pos1, pos2, score;
	msp_t *mp;
	int f;
	int pm;

	/* print the each maximal segment pair */
	for (mp=MSP_TAB_FIRST(mt); MSP_TAB_MORE(mt,mp); mp=MSP_TAB_NEXT(mp)) {
		pos1 = mp->pos1;
		pos2 = mp->pos2;
		len = mp->len;
		score = mp->score;
		f = mp->filter;
		pm = percent_match(sf1, sf2, len, pos1, pos2);
		print_block(score/10, pos1+1, pos1+len, pos2+1, pos2+len, f, pm);
	}

	if (sf1 == sf2)
		print_block(0, 1, SEQ_LEN(sf1), 1, SEQ_LEN(sf1), 0, 100);
}

void msp_write(msp_table_t *mt)
{
	int len, pos1, pos2, score;
	msp_t *mp;
	int f;
	int i;

	for (i = 0; i < mt->num; ++i){
		mp = &mt->msp[i];
		len = mp->len;
		pos1 = mp->pos1;
		pos2 = mp->pos2;
		score = mp->score;
		f = mp->filter;
		printf("%d %d %d %d %d\n", len, pos1, pos2, score, f);
	}	
}
